// Code partially copyright Edd Dawson 2007
//
// Boost Software License - Version 1.0 - August 17th, 2003
//
// Permission is hereby granted, free of charge, to any person or organization
// obtaining a copy of the software and accompanying documentation covered by
// this license (the "Software") to use, reproduce, display, distribute,
// execute, and transmit the Software, and to prepare derivative works of the
// Software, and to permit third-parties to whom the Software is furnished to
// do so, all subject to the following:
//
// The copyright notices in the Software and this entire statement, including
// the above license grant, this restriction and the following disclaimer,
// must be included in all copies of the Software, in whole or in part, and
// all derivative works of the Software, unless such copies or derivative
// works are solely in the form of machine-executable object code generated by
// a source language processor.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
// SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
// FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.


#include "libmesh_config.h"
#include "print_trace.h"

#if defined(LIBMESH_HAVE_GCC_ABI_DEMANGLE) && defined(LIBMESH_HAVE_GLIBC_BACKTRACE)

#include <iostream>
#include <string>
#include <execinfo.h>
#include <cxxabi.h>
#include <cstdlib>

namespace libMesh
{

std::string process_trace(const char *name)
{
  std::string fullname = name;
  std::string saved_begin, saved_end;
  size_t namestart, nameend;

  /**
   * The Apple backtrace function returns more information than the Linux version.
   * We need to pass only the function name to the demangler or it won't decode it for us.
   *
   * lineno: stackframeno                 address functionname + offset
   */

#ifdef __APPLE__
  namestart = fullname.find("0x");
  if (namestart != std::string::npos)
  {
    namestart = fullname.find(' ', namestart) + 1;
    saved_begin = fullname.substr(0, namestart);
  }
  else
    namestart = 0;
  nameend = fullname.find('+');
  if (nameend == std::string::npos ||
      nameend <= namestart)
    nameend = fullname.size();
  else
  {
    nameend -= 1;
    saved_end = fullname.substr(nameend, fullname.length());
  }
#else
  namestart = fullname.find('(');
  if (namestart == std::string::npos)
    return fullname;
  else
    namestart++;
  nameend = fullname.find('+');
  if (nameend == std::string::npos ||
      nameend <= namestart)
    return fullname;
#endif

  std::string type_name = fullname.substr(namestart, nameend - namestart);

  // Try to demangle now
  return saved_begin + demangle(type_name.c_str()) + saved_end;
}


std::string demangle(const char *name)
{
    int status = 0;
    char *d = 0;
    std::string ret = name;
    try { if ( (d = abi::__cxa_demangle(name, 0, 0, &status)) ) ret = d; }
    catch(...) {  }
    std::free(d);
    return ret;
}

void print_trace(std::ostream &out)
{
  void *addresses[40];
  char **strings;

  int size = backtrace(addresses, 40);
  strings = backtrace_symbols(addresses, size);
  out << "Stack frames: " << size << std::endl;
  for(int i = 0; i < size; i++)
    out << i << ": " << process_trace(strings[i]) << std::endl;
  std::free(strings);
}

} // namespace libMesh

#else

namespace libMesh
{
void libMesh::print_trace(std::ostream &) {}

std::string demangle(const char *name) {}
}

#endif
